@@ -3,7 +3,7 @@ import logging
 import os
 import pathlib
 from base64 import b64decode
-from typing import Iterable, List, Type
+from typing import List, Optional, Type
 
 import click
 
@@ -12,7 +12,6 @@ from .authenticator import BasicAuthenticator, OrgTokenAuthenticator, ProjectKey
 from .chunk import MemfaultChunk
 from .context import MemfaultCliClickContext
 from .deploy import Deployer
-from .elf import ELFFileInfo
 from .upload import (
     AndroidAppSymbolsUploader,
     BugreportUploader,
@@ -27,10 +26,10 @@ from .upload import (
     walk_files,
 )
 from .yocto import (
-    find_elf_files_from_image,
-    find_elf_files_from_non_native_packages,
-    find_symbol_files_to_upload,
-    tmp_path_from_image,
+    BitbakeBuildEnvResult,
+    PackageDebugSplitStyle,
+    get_necessary_from_bitbake_build_env,
+    process_and_upload_yocto_symbols,
 )
 
 LOG = logging.getLogger(__name__)
@@ -242,17 +241,57 @@ def upload_mar(ctx: MemfaultCliClickContext, path: str, **kwargs):
     "-i",
     required=True,
     help="""
-    The path to the image .tar/.tar.bz2/.tar.gz/.tar.xz as produced by Yocto's do_image_*.
-    The file is expected to reside in the location where Yocto produced the file
-    (tmp/deploy/images/${PACKAGE_ARCH}/images). The tool will locate the symbol files to
-    upload in the tmp/work/ directory, based on the binaries that are inside the image.
+    The path to the image .tar/.tar.bz2/.tar.gz/.tar.xz as produced by Yocto's
+    do_image_*. The file is expected to reside in the location where Yocto
+    produced the file (tmp/deploy/images/${PACKAGE_ARCH}/images). Additionally,
+    an image with the same name but ending in -dbg.tar/.tar.bz2/.tar.gz is
+    expected to reside next to the main image. To generate it alongside the
+    required elfutils-native, add the following to your main image (note that
+    the main image will not be affected except for its build dependencies):
+
+    \b
+    DEPENDS:append = " elfutils-native"
+    IMAGE_GEN_DEBUGFS = "2"
+    IMAGE_FSTYPES_DEBUGFS = "tar.bz2"
     """,
     type=click.Path(exists=True, resolve_path=True, dir_okay=False, path_type=pathlib.Path),
 )
+@click.option(
+    "--eu-unstrip-path",
+    required=False,
+    help="""
+    Path to a local eu-unstrip binary from elfutils
+    (https://sourceware.org/elfutils/).
+
+    Not necessary if running after 'source oe-init-build-env'.
+
+    If you pass --eu-unstrip-path, you must also pass --package-debug-split-style.
+    """,
+    type=click.Path(exists=True, resolve_path=True, dir_okay=False, path_type=pathlib.Path),
+)
+@click.option(
+    "--package-debug-split-style",
+    type=click.Choice([style.value for style in PackageDebugSplitStyle]),
+    required=False,
+    help="""
+    Your project's PACKAGE_DEBUG_SPLIT_STYLE. In Poky, defaults to
+    'debug-with-srcpkg'.
+
+    Not necessary if running after 'source oe-init-build-env'.
+
+    If you pass --package-debug-split-style, you must also pass --eu-unstrip-path.
+    """,
+)
 @click_option_concurrency
 @pass_memfault_cli_context
-def upload_yocto_symbols(ctx: MemfaultCliClickContext, image: pathlib.Path, **kwargs):
-    """Upload symbols for a Yocto build.
+def upload_yocto_symbols(
+    ctx: MemfaultCliClickContext,
+    image: pathlib.Path,
+    eu_unstrip_path: Optional[pathlib.Path],
+    package_debug_split_style: Optional[str],
+    **kwargs,
+):
+    """Upload symbols for a Linux Yocto build.
 
     To see a full example, take a look at the Linux SDK example project:
     https://mflt.io/yocto-symbols-image
@@ -263,42 +302,23 @@ def upload_yocto_symbols(ctx: MemfaultCliClickContext, image: pathlib.Path, **kw
         $ bitbake my-image
         $ memfault --org-token $ORG_TOKEN \\
                    --org acme-inc --project smart-sink \\
-            upload-yocto-symbols --image build/tmp/deploy/images/raspberrypi3/my-image-raspberrypi3.rootfs.tar.bz2
+            upload-yocto-symbols --image build/tmp/deploy/images/raspberrypi3/my-image-raspberrypi3.tar.bz2
     """
     ctx.obj.update(**kwargs)
 
-    def _image_elfs(image_tar: pathlib.Path) -> Iterable[ELFFileInfo]:
-        with click.progressbar(
-            find_elf_files_from_image(image_tar),
-            label="Finding binaries in image...",
-        ) as bar:
-            yield from bar
-
-    def _tmp_work_elfs(image_tar: pathlib.Path) -> Iterable[ELFFileInfo]:
-        with click.progressbar(
-            find_elf_files_from_non_native_packages(tmp_path_from_image(image_tar)),
-            label="Searching for symbols in tmp/work...",
-        ) as bar:
-            yield from bar
-
-    result = find_symbol_files_to_upload(
-        image_elfs=_image_elfs(image),
-        tmp_work_elfs=_tmp_work_elfs(image),
-    )
-    for missing_elf_info in sorted(result.missing, key=lambda i: i.path):
-        log_func = LOG.debug if missing_elf_info.path.suffix in [".ko"] else LOG.warning
-        log_func(
-            "Could not find symbol file for %s (GNU Build ID: %s)",
-            missing_elf_info.path,
-            missing_elf_info.gnu_build_id,
+    if eu_unstrip_path and package_debug_split_style:
+        build_env = BitbakeBuildEnvResult(
+            eu_unstrip_path=eu_unstrip_path,
+            package_debug_split_style=PackageDebugSplitStyle(package_debug_split_style),
         )
+    elif eu_unstrip_path or package_debug_split_style:
+        raise click.exceptions.UsageError(
+            "Parameters '--eu-unstrip-path' and '--package-debug-split-style' must be used together."
+        )
+    else:
+        build_env = get_necessary_from_bitbake_build_env()
 
-    ElfSymbolDirectoryUploader(
-        ctx=ctx,
-        file_path=image,  # Not used. TODO: refactor!
-        authenticator=ctx.create_authenticator(OrgTokenAuthenticator, BasicAuthenticator),
-        file_paths=(elf_info.path for elf_info in result.found),
-    ).upload()
+    process_and_upload_yocto_symbols(ctx, image, build_env=build_env)
 
 
 @cli.command(name="upload-mcu-symbols")
