@@ -7,7 +7,7 @@ import urllib.parse
 from abc import abstractmethod
 from concurrent.futures import ThreadPoolExecutor
 from glob import glob
-from typing import IO, Any, Dict, Generator, Optional, Tuple, Type
+from typing import IO, Dict, Iterable, List, Optional, Tuple, Type, Union
 from zipfile import is_zipfile
 
 import click
@@ -20,10 +20,15 @@ from tqdm import tqdm
 from tqdm.utils import CallbackIOWrapper
 from urllib3 import Retry
 
-from .authenticator import Authenticator
+from .authenticator import (
+    Authenticator,
+    BasicAuthenticator,
+    OrgTokenAuthenticator,
+    ProjectKeyAuthenticator,
+)
 from .context import DeltaReleaseVersion, MemfaultCliClickContext, PackageInfo
+from .elf import elf_has_debug_info, get_gnu_build_id, is_elf
 from .functools_ext import cached_property
-from .gnu_build_id import get_gnu_build_id
 
 LOG = logging.getLogger(__name__)
 
@@ -50,7 +55,8 @@ def hash_io_hexdigest(name, data: IO, chunksize=2**20) -> str:
 def check_response(response: Response):
     if response.status_code >= 400:
         raise Exception(
-            f"Request failed with HTTP status {response.status_code}\nResponse body:\n{response.content.decode()}"
+            f"Request failed with HTTP status {response.status_code}\n"
+            f"Response body:\n{response.content.decode() if response.content else '(empty)'}"
         )
 
 
@@ -59,8 +65,15 @@ def url_quote(value: str) -> str:
 
 
 class Uploader(metaclass=abc.ABCMeta):
+    authenticator_types: List[Type[Authenticator]]
+
     def __init__(
-        self, *, ctx: MemfaultCliClickContext, file_path: str, authenticator: Authenticator
+        self,
+        *,
+        ctx: MemfaultCliClickContext,
+        file_path: Union[str, os.PathLike],
+        authenticator: Authenticator,
+        **kwargs,
     ):
         self.ctx: MemfaultCliClickContext = ctx
         self.file_path = file_path
@@ -199,6 +212,12 @@ class Uploader(metaclass=abc.ABCMeta):
 
 
 class BugreportUploader(Uploader):
+    authenticator_types = [
+        ProjectKeyAuthenticator,
+        OrgTokenAuthenticator,
+        BasicAuthenticator,
+    ]
+
     def can_upload_file(self) -> bool:
         if not is_zipfile(self.file_path):
             LOG.error("%s is not a valid zip file!", self.file_path)
@@ -213,6 +232,12 @@ class BugreportUploader(Uploader):
 
 
 class CoredumpUploader(Uploader):
+    authenticator_types = [
+        ProjectKeyAuthenticator,
+        OrgTokenAuthenticator,
+        BasicAuthenticator,
+    ]
+
     def can_upload_file(self) -> bool:
         with open(self.file_path, "rb") as f:
             hdr = f.read(4).decode("ascii", errors="ignore")
@@ -228,8 +253,21 @@ class CoredumpUploader(Uploader):
         else:
             return f"{self._projects_base_url()}/coredumps"
 
+    def post_token_extra_request_data(self) -> Optional[Dict]:
+        if self.ctx.core_info:
+            return {
+                "device_serial": self.ctx.core_info.device_serial,
+            }
+        return None
+
 
 class XedUploader(Uploader):
+    authenticator_types = [
+        ProjectKeyAuthenticator,
+        OrgTokenAuthenticator,
+        BasicAuthenticator,
+    ]
+
     def can_upload_file(self) -> bool:
         return True
 
@@ -251,6 +289,12 @@ class XedUploader(Uploader):
 
 
 class MarUploader(Uploader):
+    authenticator_types = [
+        ProjectKeyAuthenticator,
+        OrgTokenAuthenticator,
+        BasicAuthenticator,
+    ]
+
     def can_upload_file(self) -> bool:
         if not is_zipfile(self.file_path):
             LOG.error("%s is not a valid zip file!", self.file_path)
@@ -273,30 +317,30 @@ class MarUploader(Uploader):
         }
 
 
-class AospSymbolUploader(Uploader):
+class ElfSymbolUploader(Uploader):
+    authenticator_types = [
+        OrgTokenAuthenticator,
+        BasicAuthenticator,
+    ]
+
     gnu_build_id: str = ""
 
-    def _is_elf(self) -> bool:
-        with open(self.file_path, "rb") as file:
-            if file.read(4) != b"\x7FELF":
+    def can_upload_file(self) -> bool:
+        if os.path.islink(self.file_path) and not os.path.exists(self.file_path):
+            return False
+
+        with open(self.file_path, "rb") as f:
+            if not is_elf(f):
+                LOG.info("%s: Not an ELF file", self.file_path)
                 return False
 
-        return True
-
-    def _get_gnu_build_id_and_has_debug_info(self) -> Tuple[Optional[str], bool]:
-        with open(self.file_path, "rb") as f:
             elf = ELFFile(f)
-            return get_gnu_build_id(elf), elf.has_dwarf_info()
-
-    def can_upload_file(self) -> bool:
-        if not self._is_elf():
-            LOG.info("%s: Not an ELF file", self.file_path)
-            return False
-
-        gnu_build_id, has_debug_info = self._get_gnu_build_id_and_has_debug_info()
-        if not gnu_build_id:
-            LOG.info("%s: looks like an ELF but does not contain a GNU Build ID", self.file_path)
-            return False
+            gnu_build_id, has_debug_info = get_gnu_build_id(elf), elf_has_debug_info(elf)
+            if not gnu_build_id:
+                LOG.info(
+                    "%s: looks like an ELF but does not contain a GNU Build ID", self.file_path
+                )
+                return False
 
         self.gnu_build_id = gnu_build_id
         if not has_debug_info:
@@ -325,6 +369,11 @@ class AospSymbolUploader(Uploader):
 
 
 class McuSdkElfSymbolUploader(Uploader):
+    authenticator_types = [
+        OrgTokenAuthenticator,
+        BasicAuthenticator,
+    ]
+
     def can_upload_file(self) -> bool:
         with open(self.file_path, "rb") as file:
             if file.read(4) != b"\x7FELF":
@@ -333,7 +382,7 @@ class McuSdkElfSymbolUploader(Uploader):
             file.seek(0)
 
             inspector = BuildIdInspectorAndPatcher(file)
-            if not inspector.elf.has_dwarf_info():
+            if not elf_has_debug_info(inspector.elf):
                 LOG.info("%s: looks like an ELF but it has no .debug_info", self.file_path)
                 return False
 
@@ -386,6 +435,11 @@ class McuSdkElfSymbolUploader(Uploader):
 
 
 class ReleaseArtifactUploader(Uploader):
+    authenticator_types = [
+        OrgTokenAuthenticator,
+        BasicAuthenticator,
+    ]
+
     def can_upload_file(self) -> bool:
         # There are no restrictions for the type of binary a customer uses for
         # a release
@@ -438,21 +492,35 @@ class ReleaseArtifactUploader(Uploader):
         }
 
 
-class DirectoryUploader(Uploader):
+def walk_files(toplevel: str) -> Iterable[str]:
+    for root, _dirs, files in os.walk(toplevel):
+        for file in files:
+            yield os.path.join(root, file)
+
+
+class ParallelUploader(Uploader):
     uploader_cls: Type[Uploader]
 
-    def can_upload_file(self) -> bool:
-        return os.path.isdir(self.file_path)
+    def __init__(
+        self,
+        *args,
+        file_paths: Iterable[Union[str, os.PathLike]],
+        **kwargs,
+    ):
+        super().__init__(*args, **kwargs)
+        self.file_paths = file_paths
 
-    def _walk_files(self, toplevel: str) -> Generator[str, Any, Any]:
-        for root, _dirs, files in os.walk(toplevel):
-            yield from map(lambda file: os.path.join(root, file), files)
+    def can_upload_file(self) -> bool:
+        if os.path.islink(self.file_path) and not os.path.exists(self.file_path):
+            return False
+
+        return os.path.isdir(self.file_path)
 
     def upload(self, *, progressbar=True) -> bool:
         queued_uploaders = []
         did_upload = False
 
-        for file_path in self._walk_files(self.file_path):
+        for file_path in self.file_paths:
             uploader = self.uploader_cls(
                 ctx=self.ctx, file_path=file_path, authenticator=self.authenticator
             )
@@ -479,11 +547,20 @@ class DirectoryUploader(Uploader):
         raise NotImplementedError
 
 
-class AospSymbolDirectoryUploader(DirectoryUploader):
-    uploader_cls = AospSymbolUploader
+class ElfSymbolDirectoryUploader(ParallelUploader):
+    uploader_cls = ElfSymbolUploader
+    authenticator_types = [
+        OrgTokenAuthenticator,
+        BasicAuthenticator,
+    ]
 
 
 class ProguardMappingUploader(Uploader):
+    authenticator_types = [
+        OrgTokenAuthenticator,
+        BasicAuthenticator,
+    ]
+
     def __init__(
         self,
         *,
@@ -491,8 +568,9 @@ class ProguardMappingUploader(Uploader):
         file_path: str,
         authenticator: Authenticator,
         package_info: PackageInfo,
+        **kwargs,
     ):
-        super().__init__(ctx=ctx, file_path=file_path, authenticator=authenticator)
+        super().__init__(ctx=ctx, file_path=file_path, authenticator=authenticator, **kwargs)
         self.package_info = package_info
 
     def can_upload_file(self) -> bool:
@@ -511,6 +589,11 @@ class ProguardMappingUploader(Uploader):
 
 
 class AndroidAppSymbolsUploader(Uploader):
+    authenticator_types = [
+        OrgTokenAuthenticator,
+        BasicAuthenticator,
+    ]
+
     def can_upload_file(self) -> bool:
         if not os.path.isdir(self.file_path):
             LOG.error(
@@ -581,8 +664,11 @@ class AndroidAppSymbolsUploader(Uploader):
         )
         if not os.path.isdir(cmake_path_out):
             return False
-        sym_dir_uploader = AospSymbolDirectoryUploader(
-            ctx=self.ctx, file_path=cmake_path_out, authenticator=self.authenticator
+        sym_dir_uploader = ElfSymbolDirectoryUploader(
+            ctx=self.ctx,
+            file_path=cmake_path_out,
+            authenticator=self.authenticator,
+            file_paths=walk_files(cmake_path_out),
         )
         return sym_dir_uploader.upload()
 
